/*------------------------------------------------------------------------

 Name:     ASCII4.C
 
 Purpose:  This is an IT work logfile helper...
  - this is the work horse module ...caller of XML and SQL processing
  - replace unix newlines with industry standard newlines
  - take longfile names
  - accept drag n drop'd files
  - To DO: XML functions should have a different module ...like  SQL does. (SQL support was an afterthought)

----------------------------------------------------------------------*/
#include <stdarg.h>
#include "LogFile.h"

#define DEBUG 0

#define FILE_EXIST(name)  (_access((name),0) != -1)  //more readable macro

#define MAX_DATA_RUN	300  /* Runaway detect: Dont let data between tags exceed this many chars */
#define MAX_XML_RUN_NO_SYMBOLS 700  /* Runaway detect: Don't allow XML 'inTags' state persist 
									** when last token ('<' or '>') hasn't been seen for XXX chars */

#define MAX_LMARGIN  550 /* left margin pushed farther than this? probably an error...*/

#define MAX_LOOKAHEAD_XML  700

#define TRANSLATED_OPAQUE (char)(169) //<--Copyright symbol will appear in html if we've miscoded

/*----------------------------------------------------------------------

  Reading only 1 byte at a time is killing the visibility we need to make decisions 
  on tag stop/start handling.

  Solution:  a proficient/transparent buffering system...to keep an array stocked from the file.
  Ideal: whole file in an array
  Required: 1000 bytes buffered look-ahead at all times

  Desired Features:
  - automatic reloads when < 1000 bytes remains in the look-ahead buffer (up to EOF)
  Debug: use very small buffers and observe results--expand buffer size for release

*-------------------------------------------------------------------------*/

char g_readBuff[READ_BUFF_SIZE];
int g_CurOffs = 0;
int g_LastByteOffs = -1;
int g_currFileSize=0;
int g_currFileOffs=0;

/*----------------------------------------------------------------------
--
-- getBufferBytesRemain()
--
-- 
--
----------------------------------------------------------------------*/
int getBufferBytesRemain(void)
{
	return max(g_LastByteOffs-g_CurOffs +1, 0);
}


/*----------------------------------------------------------------------

  _fmt_strnicmp()
  
	no match: return 0
	match: retunr number of scanned bytes

  On entry: str points to buffer, pat to the pattern, len is buffer len

*-------------------------------------------------------------------------*/
int	_fmt_strnicmp (char *str, char *pat, int len)
{
int sch=1,pch=2, scanned =0;

	if(len) 
	{
		do
		{
		  sch = tolower(*str++);
		  pch = tolower(*pat++);
		  scanned++;
		  
		  while(sch < 33 && scanned < len && pch != sch) // some unprintable or CR/LF?
		  {
			  sch = tolower(*str++);
			  scanned++;
		  }
		  if(! *pat) //break if pattern runs out
			  break;

		} while (scanned < len  && sch && pch && sch == pch);
	}
	if(pch == sch)
		return scanned;
	return(0);
}



/*----------------------------------------------------------------------

  fmtBuffer()  
	
	strip and / or replace characters as desired:
	- transform '&gt;' sequences into '>'
	- transform '&lt;' sequences into '<'
	- remove excess whitespace
	- remove all tabs
  
	Note 1: translating '&gt;' to tokens (>) enhances code functionality and readabiliy
	in parseFile()...but we'll have to undo later if user desires untranslated opaque strings.
	Solution is to enter an encoding byte to flag the system that we translated tokens,
	for later restoration.

  Note 2: fmtBuffer() often returns fewer bytes than len, due to its compression 
         nature of removing excess whitespace, tabs, etc..

*-------------------------------------------------------------------------*/
int fmtBuffer(char *pBuff, int len)
{
int src=0,dest =0;
char curCh, lastCh, nextCh;
int scanned;

	while(src < len)
	{
		lastCh =0;
		if(src > 0)
			lastCh = pBuff[src-1]; 
		curCh = pBuff[src++];
		nextCh = pBuff[src];

		switch(curCh)
		{
			/* Dang! what if a'&gt;' sequence is broken by a buffer boundary? */
			/* It should never happen.  LoadBuff passes us the whole buffer from start,
			** and all the bytes to the end of the bytes read*/
			/* What if a'&gt;' sequence is broken by a CR/LF? (it happens) 
			** _fmt_strnicmp() implementation will now handle */
			case '&':
				if(len - src > 2) /* This code looks X bytes ahead of current offset */
				{
					/* a literal may be encountered */
					scanned = _fmt_strnicmp(&pBuff[src], "gt;", len - src);
					if(scanned)
					{
						src += scanned;
						pBuff[dest++] = TRANSLATED_OPAQUE; /* encode buffer with translated signal */
						pBuff[dest++] = '>';
						break;
					}
					scanned = _fmt_strnicmp(&pBuff[src], "lt;", len - src);
					if(scanned )
					{
						src += scanned;
						pBuff[dest++] = TRANSLATED_OPAQUE;  /* encode buffer with translated signal */
						pBuff[dest++] = '<';
						break;
					}
				}
				pBuff[dest++] = curCh;
			break;
			/* Note:  HTML translates multiple spaces into one... i'll do same*/
			case ' ': 
				if(lastCh != ' ') /* _if_ not a sequence of many whitespaces...*/
					pBuff[dest++] = curCh;  /* ...allow this whitespace*/
			break;

			case '\t':  /* tabs don't have meaning in Html */
				if(lastCh != ' ') /* if no previous  whitespace...*/
					pBuff[dest++] = ' '; /* ..r2 with a whitespace in case tab removal would cause words to join!*/
			break;

			/* Either CR (0x0D = 13 = \r)  or LF (0x0A =10 = \n) or , or CR+LF(both \r\n); 
			** LF (0x0A) only =  Unix and Unix-like systems ...(see this lots in Logfiles)
			** Most textual Internet protocols mandate the CR+LF pair (\r\n)...
			**
			** Dealing with CRLF pairs in function parsefile() code is getting tedious, let's strip
			** it down to only one standard permitted in buffer 0x0A ('\n') only.
			**  This is probly what prompted Unix systems to goto a single char (0x0A)
			** 
			*/
			/* Note: we have no case for '\n' ... it is permitted in all instances */
            case '\r':
				if(nextCh != '\n') /* If not a CRLF pair...it is alone... */
					pBuff[dest++] = '\n';  /* ...then allow the entry, but replace with '\n'  */
				/* else ...nextCh == '\n' ... no worries about unwanted word joins  
				...so skip '\r' completely */
			break; 

			default:
				pBuff[dest++] = curCh; /* permit all other characters to the party */
			break;
		}
	}
	return dest;
}

/*----------------------------------------------------------------------

  LoadBuffer()
  
	return TRUE on success....

  On entry, g_CurOffs must point to next, _unread_ byte in buffer, or zero
*-------------------------------------------------------------------------*/
int LoadBuffer(FILE *srcFile)
{
int FileBytesRemain = max(g_currFileSize - g_currFileOffs, 0);
int readbytes, BufferBytesRemain = getBufferBytesRemain();
int capCh=0;
int nead_read = FileBytesRemain && (BufferBytesRemain < MIN_LOOK_AHEAD) ;

	/* First thing to be cognizant of, g_CurOffs points to next, unread byte in buffer
	** Therfore, does this code do the right thing?
	**
	** BufferBytesRemain = max(g_LastByteOffs-g_CurOffs +1, 0);
	** A: yes...even if g_CurOffs is at the last offset, it shows one byte remains...correct.
	*/

	/* Next is a situation we want to handle regardless of file bytes remaining, for 
	** safety but also code clarity ...*/
	/* in the danger zone?...compromising the buffer look-ahead? 
	** ...or just need to set up for next read?
	** ... Then shift Bytes left */

	if(BufferBytesRemain)
	{
		if(g_CurOffs > MAX_OFFSET || (nead_read && g_CurOffs)) 
		{
			if(g_CurOffs > BufferBytesRemain)
			{
				/* memcpy() is faster than memmove(), but cannot be used unless u 
				** can guarantee write & read portions of buffer do not overlap*/
				memcpy(g_readBuff, &g_readBuff[g_CurOffs], BufferBytesRemain );
			}
			else{ /* If compression induced by fmtBuffer exceeeds 50%, then code will
				** come here on subsequenst read, and g_CurOffs < BufferBytesRemain...tested.
				** Unprobable, but possible if a file has excessive whitespace, etc. */
				memmove(g_readBuff, &g_readBuff[g_CurOffs], BufferBytesRemain );
			}
			capCh = FileBytesRemain? 0 : EOF;
			memset(&g_readBuff[BufferBytesRemain], capCh, READ_BUFF_SIZE-BufferBytesRemain);
			g_LastByteOffs = BufferBytesRemain - 1;
			g_CurOffs = 0;
		}
	}
	
	/* per requirements...maintain a 1000 byte look-ahead buffer */
	if( nead_read )
	{
		/* ....then read! */ 
		readbytes = min((READ_BUFF_SIZE-BufferBytesRemain), FileBytesRemain);
		readbytes = max(0,readbytes);
		if(readbytes)
			readbytes = fread(&g_readBuff[BufferBytesRemain], 1, readbytes, srcFile);
		g_currFileOffs = ftell(srcFile);
		
		BufferBytesRemain += readbytes;
		/* We place fmtBuffer() here to insure we format any sequence that straddled the end of 
		** last buffer load.  We are then, rechecking/formatting up to MIN_LOOK_AHEAD bytes each read */
		/* fmtBuffer() may return fewer bytes than read, to to its compression nature */
		BufferBytesRemain  = fmtBuffer(g_readBuff, BufferBytesRemain);
		if(BufferBytesRemain < READ_BUFF_SIZE ) 
		{
			FileBytesRemain = max(g_currFileSize - g_currFileOffs, 0);
			capCh = FileBytesRemain? 0 : EOF;
			memset(&g_readBuff[BufferBytesRemain], capCh, READ_BUFF_SIZE-BufferBytesRemain);
		}
		g_LastByteOffs = BufferBytesRemain - 1;
		g_CurOffs = 0;
	}
	return BufferBytesRemain > (0);
}

/*----------------------------------------------------------------------

  getFileSize()
  
*-------------------------------------------------------------------------*/
int getFileSize(FILE *srcFile)
{
int flen, curPos;

	if(!srcFile)
		return 0;

	curPos = ftell(srcFile);
    fseek(srcFile, 0, SEEK_END);
    flen = ftell(srcFile);
    fseek(srcFile, curPos, SEEK_SET);
	return flen;
}

/*----------------------------------------------------------------------

  initBufferingSystem()
  
*-------------------------------------------------------------------------*/
static int initBufferingSystem(FILE *srcFile)
{
	g_LastByteOffs = -1;
	g_CurOffs = g_currFileOffs=0;
	g_currFileSize = getFileSize(srcFile);
	fseek(srcFile, 0, SEEK_SET);
	if(LoadBuffer(srcFile))
		return g_currFileSize;
	return 0;
}



/*----------------------------------------------------------------------

  SpinTheMessageLoop()

  Ck for Windows messages during long ops, such as file read-write that we do.
  You can tell when windows programmers forget this...when the running app
  seems to suck all resources & permit nothing else to happen until it's done.
  
*-------------------------------------------------------------------------*/
BOOL SpinTheMessageLoop(HWND hWnd)
{
MSG msg;

	if(hWnd); //unused

	if(PeekMessage((LPMSG)&msg,NULL,0,0,PM_REMOVE)) //hWnd or NULL
	{
		TranslateMessage (&msg);
		DispatchMessage (&msg);
	}
	return TRUE;
}

/*----------------------------------------------------
  extract_path_and_filename()
  
	Extract path and filename without modifying inbound string

 ----------------------------------------------------*/
void extract_path_and_filename(char *pszIn, char *pszPath, char *pszFname)
{
char  *ptr;

	strcpy(pszPath, pszIn);

	ptr=pszPath + strlen( pszPath );
	while(*ptr != '\\' && *ptr != '/' && *ptr != ':' && ptr>pszPath)
		ptr--;
	strcpy(pszFname,ptr+1); //copy part just past the '\\'  

	*(ptr+1) ='\0'; //filename gone! (string now ends with trailing slash '\'
}


static int lmargin = 0;
static int in_paragraph = FALSE;
static int bold =0;
static int newline = 1;

struct tagStatus{
int inTags;
int state;
int prev_state;
}tag;

enum tag_states
{
NO_TAGS,
INSIDE_START,
POST_START,
INSIDE_END,
POST_END,
COMMENT,
IN_DATA_RUN
};


/*----------------------------------------------------
  
	setTagState()
  
 ----------------------------------------------------*/
void setTagState(int state)
{
	tag.prev_state = tag.state;
	tag.state = state;
	tag.inTags = !!tag.state;
}

/*----------------------------------------------------
  
	getTagState()
  
 ----------------------------------------------------*/
int getTagState(void)
{
	if(tag.inTags)
		return tag.state;
	return NO_TAGS;
}

/*----------------------------------------------------
  
	writeMargin()
  
 ----------------------------------------------------*/
static void writeMargin(FILE *outfPtr)
{
char indentCommand[100];
int spaces,iii=0;

	if(plain_text)
	{
		spaces= lmargin/5;
		while(spaces--)
		{
			indentCommand[iii++]=' ';
		}
		indentCommand[iii++]='\0';
	}
	else
		sprintf(indentCommand,"\r\n<p style=\"margin-top:0;margin-bottom:0;margin-left:%dpt\">",lmargin);

	fwrite(indentCommand, 1, strlen(indentCommand), outfPtr);
	in_paragraph = TRUE;
}

 /*----------------------------------------------------
  
	writeCR()


 ----------------------------------------------------*/
static void writeCR(FILE *outfPtr)
{
	newline = TRUE;
	
	if(!plain_text)
	{
		if(in_paragraph)
		{
			fwrite("</p>\r\n", 1, 6, outfPtr);
			in_paragraph = FALSE;
		}
		else
		{
			fwrite("<br>\r\n", 1, 6, outfPtr);
		}
	}
	else
		fwrite("\r\n", 1, 2, outfPtr);

	if(getTagState() && lmargin)
	{
		writeMargin(outfPtr);
	}
}



#define MARGIN_GRANULARITY 15
void popMargin(void)
{
	lmargin -= MARGIN_GRANULARITY;
	if(lmargin < 0)
		lmargin = 0;
}
void pushMargin(void)
{
	lmargin += MARGIN_GRANULARITY;
	if(lmargin > MAX_LMARGIN)
		lmargin -= MARGIN_GRANULARITY;
}


/*----------------------------------------------------------------------

- isXML() -- really XML? 

  We can only display right if it is really XML...and XML follows certain rules... 
  Note that MSIE does a good job of displaying XML.  Did I really say that?
 
*-------------------------------------------------------------------------*/
int isXML(const char *inStr, int len )
{
	/* pretty much a simple token check... if tokens exist...this is XML */
	if(__strnistr(inStr, "/>", len) || __strnistr(inStr, "</", len) )
	{
		return TRUE;
	}
	return FALSE;
}

/*----------------------------------------------------------------------

	killBold()   

*-------------------------------------------------------------------------*/
void BoldOn(FILE *outfPtr)
{
	if(!bold && !plain_text) 
	{
		fwrite("<b>", 1, 3, outfPtr); 
		bold = TRUE;
	}
}

/*----------------------------------------------------------------------

	BoldOff()   

*-------------------------------------------------------------------------*/
void BoldOff(FILE *outfPtr)
{
	if(bold && !plain_text) //if bolding on, turn it off
	{
		fwrite("</b>", 1, 4, outfPtr); 
		bold = FALSE;
	}
}

/*----------------------------------------------------------------------

	killTagState()   

*-------------------------------------------------------------------------*/
void killTagState(FILE *outfPtr)
{
	setTagState(NO_TAGS);
	BoldOff(outfPtr);
	if(lmargin)
	{
		while(lmargin)
			popMargin();
		writeMargin(outfPtr);
		writeCR(outfPtr);
	}
}

/*----------------------------------------------------------------------

	writeToken()   

  Clean token writer '<' or '>' or literals ('&lt;' or '&gt;')

  Ease of implementation per Andrew's suggest: "translate opaque strings" is optional
*-------------------------------------------------------------------------*/
void writeToken(FILE *outfPtr, int ch, int *is_xlated_opaque, int xlate)
{
int write_literal = *is_xlated_opaque && !xlate;
char *tokens[6]={{"&lt;"},{"&amplt;"},{"&gt;"},{"&ampgt;"},{"ERR_XL"},{"ERR_XL"}};
char *pt_tokens[6]={{"<"},{"&lt;"},{">"},{"&gt;"},{"ERR_XL"},{"ERR_XL"}};
char *ptr=NULL;  
char index=0;

	switch(ch)
	{
		case '<':
			index = 0;
		break;
		case '>':
			index = 2;
		break;
		default:
			index = 4; /* Code should never get here. But if it does, make error visible/traceable 'ERR_XL'*/
		break;
	}
	if(write_literal)
		index++;
	ptr = tokens[index];
	if(plain_text)
		ptr = pt_tokens[index];

	fwrite(ptr, 1, strlen(ptr), outfPtr);  //escape:  write html '&lt;'
	*is_xlated_opaque = 0; /*turn off the flag for caller*/
}


/*------------------------------------------------------------------------
 
 parsefile()
 
 
 History:   Date      Reason
             30JUL96  Created 
			 July-Aug 2009 modified for logFile handling

--------------------------------------  
Every '<' and '>' token is a *potential* margin adjust.
'<' push indent
'</' pop indent
  
Want indent/margin neat like this:
   <someTag>						(start block)
	 <someTag>						(start child block)
		<someTag></someTag>	     ---empty tag type 1
		<someTag/>				 ---empty tag type 2 ('shorthand')
		<someTag>someData</someTag>  --data---
	 </someTag>			        	(end child block)
   </someTag>						(end block)

------------------------------------------------------------------------*/
BOOL parsefile(HWND hwnd, FILE *infPtr, char *szInFileName)
{

static FILE *outfPtr;
static int  curCh, lastCh=0, nextCh;
static long dwChCount=0, fLen, dwCharCountSinceTag = 0, bytesRead, SQLbytesRead, inDataCount;
static long last_g_CurOffs=0;
static int done = FALSE, tries;
static char szOutFileName[MAX_PATH], *ptr;
static char szPath[MAX_PATH],szFname[MAX_PATH];
static int needNewLine;
static int TagState =0, xlated_opaque =0 ;
static char *pExt=NULL;

	if(!infPtr)
	{
		//MessageBox(NULL, "File pointer is NULL", "ERROR", MB_OK );
	    return FALSE;
	}

	/* get file len & init buffering system */
    fLen = initBufferingSystem(infPtr);
	if(!fLen)
	{
		MessageBox(NULL, "Empty File", "ERROR", MB_OK );
	    return FALSE;
	}

	extract_path_and_filename(szInFileName, szPath, szFname);
	
	/* Create the html fname */
	ptr=szFname + strlen( szFname );
	while(*--ptr != '.' && ptr>szFname);
	*ptr ='\0'; //extension gone!
	if(plain_text)
		pExt=".txt";
	else
		pExt=".html";

	sprintf(szOutFileName, "%s%s%s", szPath, szFname, pExt);

	/* Set OutFile.html name.
	** We don't overwrite files.  And we don't bug users for a filename.  
	** So how deal?  Like this... 
	** Ensure this file name doesn't exist...tweak name until it doesn't exist (get it?) */
	tries = 0;
	while(FILE_EXIST(szOutFileName)) //as long as a file exists...
		sprintf(szOutFileName, "%s%s%d%s", szPath, szFname, tries++, pExt); //tweak name with increment

    /*OPEN *BINARY* html file for write.*/
    outfPtr = fopen(szOutFileName, "w+b");    //append would be "a+"
    if(!outfPtr)
    {
		MessageBox(hwnd, "Couldn't open Outfile", "ERROR", MB_OK );
		fclose(infPtr);
        return FALSE;
    }

	/* Set up HTML header */
	if(!plain_text)
	{
		ptr = "<HTML>\r\n<body style=\"font-family:verdana;color:black;font-size:12px\">\r\n\r\n";
		fwrite(ptr, 1, strlen(ptr), outfPtr); 
	}

	lmargin = 0;
	g_CurOffs = 0;
	newline = TRUE;
	setTagState(NO_TAGS);
	lastCh=0;
	inDataCount = dwChCount=0;
	done = in_paragraph = FALSE;
	xlated_opaque =0;

	//Parse the file, write to out file
	while(!done)
    {
		/* Always have last, cur, and next char now */
		curCh = g_readBuff[g_CurOffs];
		last_g_CurOffs = g_CurOffs ;
		nextCh = g_readBuff[g_CurOffs+1];

		if(!(dwChCount % 1000)) /*every 1000 -- not every char -- that would kill performance */
		{
		/* Time to update progress bar and ck Windows OS mssgs */
		int bar_pos = (int)(((__int64)dwChCount*(__int64)PROGRESS_MAX_RANGE)/(__int64)fLen);
			/* Update progress bar */
			SendMessage(hWndProgressBar, PBM_SETPOS, (WPARAM)bar_pos, (LPARAM)0);
			SpinTheMessageLoop(hwnd); /* Ck OS messages*/
		}

		if( gQuitMssg )
			break;

        switch(curCh)
        {
			case TRANSLATED_OPAQUE: /* our encoding char */
				xlated_opaque = 1;  /* Set flag ... next tag token is translated from opaque string*/
				curCh = lastCh; /* trick sequence at bottom of loop -- like this encode character never happened */
			break;

			case '<':  /* &lt; -- 'less than' */
				needNewLine = !newline;

				/* XML: '<' is a new element start...other usage of '<' is illegal in XML.
				** But -- in a log file -- there's no guarantee that tokens'<' '>' will signal
				** XML, anymore than 'select' guarantees presence of SQL.  So, we have a lot 
				** of rigorous cks & balances to code 	*/
				dwCharCountSinceTag = 0;
				BoldOff(outfPtr);
				
				/* Before we can make assumptions about what this char means to us, cks must be made
				*...what is the current tag state?  What is the margin? ...etc
				*...If we are in a data run, this char is illegal data...*/

				/* Also, were not allowing natural CRs while in tags...must write our own. 
				*/
				switch(nextCh)
				{
					case '/':  /* ---- we have a </stopTag> token ---- */
						popMargin();
						if(getTagState() == IN_DATA_RUN)	
							needNewLine = FALSE;
						/* Occasionally, in fragmented XML log, we could get a </stop tag>
						** although not in a tag state */
						if(getTagState() == NO_TAGS)	
							needNewLine = FALSE;
						setTagState( INSIDE_END );
					break;

					case '!':  	/* <!--Comment?-->  */
						if(g_readBuff[g_CurOffs+2] == '-')
							setTagState( COMMENT );
						else { /* <!DOCTYPE note [ or CDATA... (hold for future use) */  }
					break;

					default:  /* ----  assume is <start tag> token ----*/
						/* 
						** The </stop tag> will do the popMargin.  If </stop tag> appears on
						** current line or end of block, current code will handle the margin pop
						** the right way, every time for the desired effect... as long as every
						** <start tag> does unconditional pushMargin*/
						if(getTagState() == NO_TAGS)
						{
							if(	!isXML(&g_readBuff[g_CurOffs+1], MAX_LOOKAHEAD_XML ) )
							{
								needNewLine = FALSE;
								break;
							}
						}
						if(needNewLine)
							writeCR(outfPtr);
						needNewLine = FALSE;
						pushMargin();  /* .. unconditionally: start tag or start block.. no diff */
						setTagState( INSIDE_START );
					break;
				}
				
				if(	needNewLine )
					writeCR(outfPtr);

				writeToken(outfPtr, curCh, &xlated_opaque, translate_opaque_strings);
				newline = 0;

				switch(getTagState())
				{
					case COMMENT:  		
					break;
					case INSIDE_START:  
					break;

					case INSIDE_END: 	
					break;

					case POST_START:
					case IN_DATA_RUN:
						inDataCount=0;
						setTagState( INSIDE_END );
					break;
					
					case POST_END:
						setTagState( INSIDE_START );
					break;

					case NO_TAGS:
					break;
										
					default:
					break;
				}
				
			break;

			case '>':  /* &gt; -- 'greater than' */
				dwCharCountSinceTag = 0;

				if(lastCh == '/')
				{
					/* This is -- <empty tag/> (short hand) token ---
					*  ...there won't be a matching tag name in the stack
					*  It's push / pop occurs in one tag...push was already done, so...
					*/
					popMargin();
					setTagState( POST_END );
				}
				writeToken(outfPtr, curCh, &xlated_opaque, translate_opaque_strings);
				newline = 0;

				switch(getTagState())
				{
					case NO_TAGS:
					break;

					case INSIDE_START:
						setTagState( POST_START );
						/* Under normal circumstances, data will now follow */
					break;

					/* Next 2 cases should not happen ...but if they do...*/
					case POST_START:
					case IN_DATA_RUN:  
						killTagState(outfPtr);  /* Illegitimate condition...quit XML  */
					break;
					
					case INSIDE_END:
					case COMMENT:
					case POST_END: /* <-- should not see this case!*/
						setTagState( POST_END );
						/* See if this is an XML stop
						** Really, we should not take lmargin into account..XML could be badly malformed
						** ...lmargin way off ...*/
						/* Let's see if this is a false stop...does XML begin again soon? */
						if(__memnchr(&g_readBuff[g_CurOffs+1], '<', 20 ) == -1)
						{
							killTagState(outfPtr);  /* No, no tags soon...set NO_TAGS + cleanup */
#if DEBUG
							fprintf( outfPtr, "std stop");
#endif //DEBUG
						}
					break;
										
					default:
					break;
				}
			break;
		
				/* Toss CRs/LFs until i want (note no 0x0D case..now stripped by fmtBuffer()*/
				case '\n':
 				switch(getTagState() )
				{
					case NO_TAGS:
						writeCR(outfPtr);
					break;
					case POST_END:
						/* no indent and no start token soon ? */
						if(!lmargin && __memnchr( &g_readBuff[g_CurOffs+1], '<', 40) == -1)
						{
							killTagState(outfPtr);  /* Yes. Done with XML run */
#if DEBUG
							fprintf( outfPtr, "stop for CR");
#endif //DEBUG
						}
					case IN_DATA_RUN:
						/* Just like standard apps: strip out CR/LFs and replace w/ spaces...*/
						if(nextCh != '<' )
						{
							fputc(' ',outfPtr);
							newline = 0;
						}
					break;
					default:
					break;
				}
            break;

			case EOF:
				done = TRUE;
				#if DEBUG
					sprintf(szPath,"EOF at: dwChCount: %ld  g_currFileSize: %ld", dwChCount, g_currFileSize);
					MessageBox(NULL, szPath, "INFO", MB_OK );
				#endif //DEBUG
			break;


            default:  /* any other char ... letters, numbers...*/
				SQLbytesRead = 0;
				switch(getTagState())
				{
					case NO_TAGS:
						/* Ck for SQL */
						if(isalpha( curCh ) && isSQL(&g_readBuff[g_CurOffs], lastCh))
						{
							killTagState(outfPtr); /* perform any needed cleanup */
#if DEBUG
							fprintf( outfPtr, "stop for SQL");
#endif //DEBUG
							SQLbytesRead = strFmtSQL(infPtr, outfPtr, &curCh, dwChCount, fLen);
							newline = TRUE;
#if DEBUG
							if(curCh != ';')	
								MessageBox(NULL, "strFmtSQL() is doing the wrong thing", "CODE ERROR", MB_OK );
#endif //DEBUG
						}
					break;

					case INSIDE_START:
					case INSIDE_END:
					case COMMENT:
					case POST_END:
						BoldOff(outfPtr);
					break;

					case POST_START:
						setTagState(IN_DATA_RUN);
						BoldOn(outfPtr);
					break;
					case IN_DATA_RUN:
					break;
				
					default:
					break;
				}
				if(!SQLbytesRead)
				{
					fputc(curCh,outfPtr);
					newline = 0;
				}
				SQLbytesRead = 0;
            break;
        }
        lastCh = curCh;
		bytesRead = g_CurOffs - last_g_CurOffs;
		if(!bytesRead)
		{
			bytesRead = 1;
			g_CurOffs++;
		}
		dwChCount += bytesRead;
		dwCharCountSinceTag += bytesRead;
		
		/* Note about LoadBuffer()... it is now the de-facto signal for quitting.  Since we 
		** format the buffer and --in effect-- compress it, (dwChCount >= fileLength) is no longer 
		** reliable for determining EOF.
		*/
		if(!LoadBuffer(infPtr))
		{
#if DEBUG
			sprintf(szPath,"LoadBuffer() out of Bytes: dwChCount: %ld  g_currFileSize: %ld", dwChCount, g_currFileSize);
			MessageBox(NULL, szPath, "INFO", MB_OK );
#endif //DEBUG
			done = TRUE;
			break;
		}

#if DEBUG
		switch(TagState)
		{
			case 0:
				if(getTagState())
				{
					fprintf( outfPtr, "<br><b>&lt;-## - BEGIN XML - ##-&gt;</b><br>");
					TagState = 1;
				}
			break;
			case 1:
				if(!getTagState())
				{
					fprintf( outfPtr, "<br><b>&lt;-## - END XML - ##-&gt;</b><br>");
					TagState = 0;
				}
			break;
		}
#endif //DEBUG
		if(getTagState() == IN_DATA_RUN )
		{
			inDataCount += bytesRead;
			if(inDataCount > MAX_DATA_RUN) /* 150 may be too chincy */
			{
				/* We're probly not actually in tags anymore...runaway mode */
				killTagState(outfPtr);  
#if DEBUG
				fprintf( outfPtr, "stop excess data");
#endif //DEBUG
				inDataCount = 0;
			}
		}
		else
			inDataCount=0;

		if(bold && getTagState() != IN_DATA_RUN)
		{
			BoldOff(outfPtr);
			/* Solves old Code BUG ... how did we get here?  This sequence did it:
			** "<[someTokenName{ variedstring =
			**  <somevalue> }]>".  
			**	The ' }]' sequence was in bold, per logic.
			** But the '>' arrival caused tagstate to change...we neglected bold off
			*/
#if DEBUG
			MessageBox(hwnd, "bold && getTagState() != IN_DATA_RUN", "ERROR", MB_OK );
#endif //DEBUG
		}

		/* This code is kind of clumbsy, but we do need a runaway breakout */
		if(getTagState() &&  dwCharCountSinceTag > MAX_XML_RUN_NO_SYMBOLS  && 
				getTagState() != IN_DATA_RUN)
		{
#if DEBUG
			fprintf( outfPtr, "stop %d count", MAX_XML_RUN_NO_SYMBOLS);
#endif //DEBUG

			killTagState(outfPtr);
		}
		if(dwChCount >= fLen) /* we're done and we don't know it? */
		{
#if DEBUG
			sprintf(szPath,"dwChCount >= fLen at: dwChCount: %ld  g_currFileSize: %ld", dwChCount, g_currFileSize);
			MessageBox(NULL, szPath, "INFO", MB_OK );
#endif //DEBUG
            done = TRUE;
		}
    }

	/* Set status bar to max */
	SendMessage(hWndProgressBar, PBM_DELTAPOS, (WPARAM)(PROGRESS_MAX_RANGE), (LPARAM)0);

	/* Give Key Word frequency stats - development use only */
	/* printSQLkwFreq(outfPtr);  */


	/* Write HTML footer */
	if(!plain_text)
		ptr = "</p>\r\n</body>\r\n</HTML>\r\n"; 
	else
		ptr = "\r\n\r\n"; 

	fwrite(ptr, 1, strlen(ptr), outfPtr); 

    fclose(outfPtr);

    return TRUE;

}



