/*------------------------------------------------------

 CHECKEXE.C

 checksum protection for executable files by: Bob Jarvis

  Bob's original version utilized CheckSum...where every byte of the executable is summed
  and stored in the program.  
  
  The concept is: when the program runs, it does a checksum check on itself - on the executable.
  If the sum doesn't match the stored value, then the program exits for safety...
  ... becuase a fail likely means the program was currupted
  during download, by a virus, or hacker-tampering.

  Great idea, but I decided to implement a 16 bit CRC
  (Cyclic Redundancy Check) in place of CheckSum routine.
  The CRC ck offers much stronger protection..it is hard to defeat.

  This version uses the public domain CRC-CCiTT CRC algorythm...more info and code at
  http://www.embeddedrelated.com/groups/msp430/show/29689.php
  ..and many places on the web

  Note:  MSVC++ never compiles the same way 2X
	ie: the CRC will be different on each compile

  This requires a seperate exe "stamping" utility.
	- Immediately after compile, run the stamp utility to store the  CRC value into the new
	LogFile.exe.  After that, the CRC ck will pass; program is protected.
	Asumption: program is created on a virus-free system.

****************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <conio.h>
#include <ctype.h>
#include <limits.h>
#include <stddef.h>

static unsigned char *CRCBuff; 

// #define TEST
#ifdef TEST
#else
	#include "LogFile.h"
#endif


static struct tagCRCblock
{
    unsigned char sMark[8];
    unsigned short saved_CRC;         
}ck = {"CHECKMK", 0};   



/*-----------------------------------------------------------------------------
NAME:	crc_ccitt_update()

Purpose:	Calculates a 16 bit CRC value on a single byte basis.

ALG: The CCITT algorithm used by this function was described by Jack W.
	Crenshaw in Embedded Systems Programming/January 1992.
	The following formula describes the CCITT CRC calculation:
		G(x) = x^16 + x^12 + x^5 + 1

Courtesy: Paul Curtis, Rowley Associates Ltd http://www.rowley.co.uk 
Source:  http://www.embeddedrelated.com/groups/msp430/show/29689.php

  Other sources:
http://www.google.com/search?hl=en&num=100&q=ccitt+crc32+c&btnG=Search&aq=f&oq=&aqi=
http://www.avrfreaks.net/index.php?name=PNphpBB2&file=viewtopic&t=28214&highlight=crc32
http://www.docklight.de/manual/index.htm?dl_calcchecksum.htm
http://www.netrino.com/Embedded-Systems/How-To/CRC-Calculation-C-Code
  
-----------------------------------------------------------------------------*/
unsigned short crc_ccitt_update(unsigned char byte, unsigned short crc)
{
	crc = (unsigned short)(crc >> 8) | (crc << 8);
	crc ^= byte;
	crc ^= (unsigned char)(crc & 0xff) >> 4;
	crc ^= crc << 12;
	crc ^= (crc & 0xff) << 5;
	return (crc);
}

/*-----------------------------------------------------------------------------
NAME:	myCRC

	Purpose:	Calculates a 16 bit CRC value of a buffer.

	Uses:  crc_ccitt_update() - per-byte CRC value from free websource
  
-----------------------------------------------------------------------------*/
unsigned short myCRC(void  *buffer, unsigned long size, unsigned short crc)
{
unsigned char *ptr = (unsigned char  *)buffer;

    while(size--)
		crc = crc_ccitt_update(*ptr++, crc);
	return (crc);
}



#ifdef TEST
/*----------------------------------------------------------------------

  getFileSize()
  
*-------------------------------------------------------------------------*/
int getFileSize(FILE *srcFile)
{
int flen, curPos;

	if(!srcFile)
		return 0;

	curPos = ftell(srcFile);
    fseek(srcFile, 0, SEEK_END);
    flen = ftell(srcFile);
    fseek(srcFile, curPos, SEEK_SET);
	return flen;
}

char fnamebuff[_MAX_PATH];
/* Use the actual windows version in release, our kludge for test */
void GetModuleFileName(char *p, char *sz, int len)
{
	strcpy(sz,fnamebuff); 
}
#endif


/*--------------------------------------------------------------

 CheckExe()

 crc protection for executable files

 by: Bob Jarvis  modified: gspears
*---------------------------------------------------------------*/
int CheckExe(char *mssg) 
{

FILE *fptr=NULL;
unsigned short  current_crc;
static char fname[_MAX_PATH];
unsigned int dwBytes;

	strcpy(mssg,"Success!");
	
	GetModuleFileName(NULL, fname, _MAX_PATH);
    fptr = fopen(fname,"rb"); //read only binary
    if(fptr == NULL)  
	{
		strcpy(mssg,"integrity Ck: Unable to open program file!\n");
        return 0;
    }
    
    /* compute CRC */
    dwBytes =  getFileSize(fptr);
    rewind(fptr);                      //goto start of file

	/* in Win32 -  malloc() takes (size_t) paramenter, which is now a 32 bit int 
	** >>>>> printf ( "sizeof(size_t): %d\n", sizeof(size_t));   --- prints '4'  <<<<
	** ... a 4GB capability!   So, there is no reason why we shouldn't be able to allocate the
	** whole filesize and make one read */

	CRCBuff = malloc(dwBytes);
	if(!CRCBuff)
	{
		strcpy(mssg, "Memory Allocation error");
		return 0;
	}

    if(fread (CRCBuff, 1, dwBytes,fptr) != dwBytes)
	{
        strcpy(mssg,"Read error: Failed self-integrity check");
        return (0);
    }
    current_crc = myCRC(CRCBuff, (dwBytes), 0xffff);

	free(CRCBuff);
    fclose(fptr);

    #ifdef TEST
        printf("\ncurrent_crc = %u, saved_CRC = %u\n\n",current_crc, ck.saved_CRC);
    #endif //TEST

    if(current_crc != ck.saved_CRC) 
	{
        strcpy(mssg, "Program failed integrity self-check...exiting for your safety...");
        return 0;
    }
    return 1;
}


#ifdef TEST
int main(short int argc, char *argv[])
{
char mssg[500];

	strcpy(fnamebuff, argv[0]);
	CheckExe(mssg);
	puts(mssg);

	getch();
	return 0;
}
#endif
